import sys
from math import exp, log

import xml.etree.ElementTree as ET
import random
import time
import getopt


class Coord:
    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return f'({self.x}, {self.y}, {self.z})'
    
    def __add__(self, other):
        return Coord(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Coord(self.x - other.x, self.y - other.y, self.z - other.z)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __ne__(self, other):
        return not self == other
    
    def __lt__(self, other):
        if self.z < other.z:
            return True
        elif self.z == other.z:
            if self.y < other.y:
                return True
            elif self.y == other.y:
                if self.x < other.x:
                    return True
        return False

    def __gt__(self, other):
        return other < self

    def __le__(self, other):
        if self.z < other.z:
            return True
        elif self.z == other.z:
            if self.y < other.y:
                return True
            elif self.y == other.y:
                if self.x <= other.x:
                    return True
        return False

    def __ge__(self, other):
        return other <= self
    
    def __hash__(self):
        return hash((self.x, self.y, self.z))

DIR_OFFSETS = {
    'N': Coord( 0, 1, 0),
    'S': Coord( 0,-1, 0),
    'E': Coord( 1, 0, 0),
    'W': Coord(-1, 0, 0),
    'U': Coord( 0, 0, 1),
    'D': Coord( 0, 0,-1),
}

DIR_OPPS = {
    'N': 'S',
    'S': 'N',
    'E': 'W',
    'W': 'E',
    'U': 'D',
    'D': 'U',
}

ADJ_OFFSETS = [
    Coord(-1, 0, 0), Coord(1, 0, 0),
    Coord(0, -1, 0), Coord(0, 1, 0),
    Coord(0, 0, -1), Coord(0, 0, 1),
]

def getLabelComplement(label):
    if label[-1] == '*':
        return label[:-1]
    else:
        return label + '*'

ATTACHMENT = 0
DETACHMENT = 1

class Event:
    def __init__(self, eventType, polyomino):
        self.eventType = eventType
        self.polyomino = polyomino

    def __str__(self):
        return f'event: type {self.eventType}, polyomino {self.polyomino}'
        
    def __eq__(self, other):
        return self.eventType == other.eventType and self.polyomino == other.polyomino
        
    def __hash__(self):
        return hash((self.eventType, self.polyomino))

class EventBucket:
    def __init__(self):
        self.events = []
        self.eventIndices = {}

    def sample(self, remove=False):
        event = random.choice(self.events)
        if remove:
            self.removeEvent(event)
        return event

    def addEvent(self, event):
        if event not in self.eventIndices:
            self.eventIndices[event] = len(self.events)
            self.events.append(event)
            return True
        else:
            print(f'WARNING attempted to add an already existing event {event}')
            return False

    def removeEvent(self, event):
        if event in self.eventIndices:
            index = self.eventIndices[event]
            if index == len(self.events) - 1:
                del self.events[index]
                del self.eventIndices[event]
            else:
                lastEvent = self.events.pop()
                self.events[index] = lastEvent
                self.eventIndices[lastEvent] = index
                del self.eventIndices[event]
            return True
        else:
            print(f'WARNING attempted to remove a non-existing event {event}')
            return False

class StochasticEventSet:
    def __init__(self, polyominoTypes, maxStrength):
        self.eventBuckets = []
        self.polyominoTypes = polyominoTypes
        self.maxStrength = maxStrength

        self.concentrationTotal = 0
        for polyominoType in polyominoTypes:
            if polyominoType.concentration > 0:
                self.concentrationTotal += polyominoType.concentration

        self.concentrationMap = {}
        self.attachmentBuckets = {}
        for polyominoType in polyominoTypes:
            if polyominoType.concentration > 0:
                bucket = EventBucket()
                self.attachmentBuckets[polyominoType] = bucket
                self.eventBuckets.append(bucket)
            concentrationRatio = polyominoType.concentration / self.concentrationTotal
            self.concentrationMap[polyominoType] = concentrationRatio
            
        self.attachmentBucket = EventBucket()
        self.eventBuckets.append(self.attachmentBucket)
        
        self.detachmentBuckets = {}
        for strength in range(maxStrength + 1):
            bucket = EventBucket()
            self.detachmentBuckets[strength] = bucket
            self.eventBuckets.append(bucket)

    def addAttachmentEvent(self, polyomino):
        event = Event(ATTACHMENT, polyomino)
        return self.attachmentBuckets[polyomino.polyominoType].addEvent(event)

    def addDetachmentEvent(self, polyomino, strength):
        event = Event(DETACHMENT, polyomino)
        return self.detachmentBuckets[strength].addEvent(event)

    def removeAttachmentEvent(self, polyomino):
        event = Event(ATTACHMENT, polyomino)
        return self.attachmentBuckets[polyomino.polyominoType].removeEvent(event)

    def removeDetachmentEvent(self, polyomino, strength):
        event = Event(DETACHMENT, polyomino)
        return self.detachmentBuckets[strength].removeEvent(event)

    def attachmentLogProbability(self, bucket, polyominoType, Gmc):
        if len(bucket.events) == 0 or polyominoType.concentration == 0:
            return -1e9
        else:
            return log(len(bucket.events) * polyominoType.concentration) - Gmc

    def detachmentLogProbability(self, bucket, strength, Gse):
        if len(bucket.events) == 0:
            return -1e9
        else:
            return log(len(bucket.events)) - strength * Gse
    
    def sample(self, forwardRate, Gse, Gmc, remove=False):
        gumbelSamples = []
        rateAny = 0
        for polyominoType in self.polyominoTypes:
            bucket = self.attachmentBuckets[polyominoType]
            lp = self.attachmentLogProbability(bucket, polyominoType, Gmc)
            rateAny += exp(lp)
            g = -log(-log(random.random()))
            gumbelSamples.append((g + lp, bucket))
        for strength in range(self.maxStrength + 1):
            bucket = self.detachmentBuckets[strength]
            lp = self.detachmentLogProbability(bucket, strength, Gse)
            rateAny += exp(lp)
            g = -log(-log(random.random()))
            gumbelSamples.append((g + lp, bucket))
        
        gumbelSamples.sort(reverse=True)

        gi = 1
        bucket = gumbelSamples[0][1]
        while len(bucket.events) == 0:
            bucket = gumbelSamples[gi][1]
            gi += 1

        rateAny *= forwardRate
        dt = -log(1 - random.random()) / rateAny
            
        return bucket.sample(remove), dt

class PolyominoType:
    def __init__(self, name, color, concentration, blocks, domains,
                 placementHint=None, dimRestrictions=None):
        self.name = name
        self.color = color
        self.concentration = concentration
        self.blocks = blocks
        self.domains = domains

        self.placementHint = placementHint or (None, None, None)
        self.dimRestrictions = dimRestrictions or (None, None, None)

        self.ComputeGlueMap()
        self.ComputePerimeter()

    def __str__(self):
        return f'Polyomino type {self.name}: blocks: {self.blocks}, domains: {self.domains}'

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def ComputePerimeter(self):
        self.perimeter = []
        for block in self.blocks:
            for add in ADJ_OFFSETS:
                nbr = block + add
                if nbr not in self.blocks:
                    if self.dimRestrictions[0] is not None:
                        if nbr.x < self.dimRestrictions[0][0]:
                            continue
                        elif nbr.x >= self.dimRestrictions[0][1]:
                            continue
                    if self.dimRestrictions[1] is not None:
                        if nbr.y < self.dimRestrictions[1][0]:
                            continue
                        elif nbr.y >= self.dimRestrictions[1][1]:
                            continue
                    if self.dimRestrictions[2] is not None:
                        if nbr.z < self.dimRestrictions[2][0]:
                            continue
                        elif nbr.z >= self.dimRestrictions[2][1]:
                            continue
                    self.perimeter.append(nbr)

    def ComputeGlueMap(self):
        self.glueMap = {}
        for block in self.domains:
            for label, direction, strength in self.domains[block]:
                nbrCoord = block + DIR_OFFSETS[direction]
                nbrDir = DIR_OPPS[direction]
                nbrLabel = getLabelComplement(label)
                nbrStrength = strength
                self.glueMap[nbrCoord] = (nbrLabel, nbrDir, nbrStrength)
        
class Polyomino:
    def __init__(self, polyominoType, offset=None):
        self.polyominoType = polyominoType
        self.offset = offset or Coord(0, 0, 0)

        self.ComputeBlocks()
        self.ComputeDomains()
        self.ComputeGlueMap()
        self.ComputePerimeter()

    def __str__(self):
        return f'Polyomino of type {self.polyominoType.name}: offset: {self.offset}'

    def __eq__(self, other):
        return self.polyominoType == other.polyominoType and self.offset == other.offset

    def __hash__(self):
        return hash((self.polyominoType, self.offset))

    def ComputeBlocks(self):
        self.blocks = set()
        for block in self.polyominoType.blocks:
            self.blocks.add(block + self.offset)

    def ComputeDomains(self):
        self.domains = {}
        for coord, domainSet in self.polyominoType.domains.items():
            self.domains[coord + self.offset] = domainSet
            
    def ComputeGlueMap(self):
        self.glueMap = {}
        for coord, domain in self.polyominoType.glueMap.items():
            self.glueMap[coord + self.offset] = domain
            
    def ComputePerimeter(self):
        self.perimeter = []
        for coord in self.polyominoType.perimeter:
            self.perimeter.append(coord + self.offset)

class PolyominoAssembly:
    def __init__(self, polyominoTypes, minBinding):
        self.polyominoTypes = polyominoTypes
        self.minBinding = minBinding
        
        self.locationMap = {}
        self.polyominoes = set()
        self.attachmentEventMap = {}
        self.bindingStrengthMap = {}

        self.attachableTypes = []
        maxStrength = 0
        for polyominoType in polyominoTypes.values():
            if polyominoType.concentration > 0:
                self.attachableTypes.append(polyominoType)
            strength = 0
            for coord, domainSet in polyominoType.domains.items():
                for domain in domainSet:
                    strength += domain[2]
            if strength > maxStrength:
                maxStrength = strength

                
        self.eventSet = StochasticEventSet(self.attachableTypes, maxStrength)

    def AddPolyomino(self, polyomino, bSeed=False):
        if self.OverlapsExistingPolyomino(polyomino):
            raise Exception(f'Attempted to add polyomino which would overlap another: {polyomino}')

        self.polyominoes.add(polyomino)
        if not bSeed:
            strength = self.GetBindingStrength(polyomino)
            self.eventSet.addDetachmentEvent(polyomino, strength)
            self.bindingStrengthMap[polyomino] = strength

        for block in polyomino.blocks:
            self.locationMap[block] = polyomino
        self.RemoveOverlappingAttachmentEvents(polyomino)
        
        testLocations = []
        for pl in polyomino.perimeter:
            if pl in self.locationMap:
                nbrPolyomino = self.locationMap[pl]
                self.UpdateBindingStrength(nbrPolyomino)
            else:
                testLocations.append(pl)

        for possPolyomino in self.GetFittingPolyominoes(testLocations):
            if not self.InAttachmentEventMap(possPolyomino):
                self.eventSet.addAttachmentEvent(possPolyomino)
                for block in possPolyomino.blocks:
                    if block not in self.attachmentEventMap:
                        self.attachmentEventMap[block] = set()
                    self.attachmentEventMap[block].add(possPolyomino)

    def RemovePolyomino(self, polyomino):
        self.polyominoes.remove(polyomino)
        
        strength = self.bindingStrengthMap[polyomino]
        self.eventSet.removeDetachmentEvent(polyomino, strength)
        del self.bindingStrengthMap[polyomino]

        for block in polyomino.blocks:
            del self.locationMap[block]

        toRemove = set()
        for pl in polyomino.perimeter:
            if pl in self.locationMap:
                nbrPolyomino = self.locationMap[pl]
                self.UpdateBindingStrength(nbrPolyomino)
            else:
                if pl in self.attachmentEventMap:
                    for nbrPolyomino in self.attachmentEventMap[pl]:
                        bKeep = False
                        for npl in nbrPolyomino.perimeter:
                            if npl in self.locationMap:
                                bKeep = True
                                break
                        if not bKeep:
                            toRemove.add(nbrPolyomino)
        for rPolyomino in toRemove:
            self.RemoveAttachmentEventForPolyomino(rPolyomino)

        for pPolyomino in self.GetFittingPolyominoes(polyomino.blocks):
            bAdd = False
            for coord in pPolyomino.perimeter:
                if coord in self.locationMap:
                    bAdd = True
                    break
            if bAdd:
                self.eventSet.addAttachmentEvent(pPolyomino)
                for block in pPolyomino.blocks:
                    if block not in self.attachmentEventMap:
                        self.attachmentEventMap[block] = set()
                    self.attachmentEventMap[block].add(pPolyomino)

    def InAttachmentEventMap(self, polyomino):
        for block in polyomino.blocks:
            break # this is a hack to get the first element of a set quickly

        if block not in self.attachmentEventMap:
            return False
        else:
            return polyomino in self.attachmentEventMap[block]
                
    def OverlapsExistingPolyomino(self, polyomino):
        for block in polyomino.blocks:
            if block in self.locationMap:
                return True
        return False

    def RemoveAttachmentEventForPolyomino(self, polyomino):
        for block in polyomino.blocks:
            self.RemoveAttachmentEventFromMap(block, polyomino)
        self.eventSet.removeAttachmentEvent(polyomino)
    
    def RemoveOverlappingAttachmentEvents(self, polyomino):
        for block in polyomino.blocks:
            self.RemoveAttachmentEventAtLocation(block)

    def RemoveAttachmentEventAtLocation(self, coord):
        if coord in self.attachmentEventMap:
            for polyomino in self.attachmentEventMap[coord]:
                for block in polyomino.blocks:
                    if block != coord:
                        self.RemoveAttachmentEventFromMap(block, polyomino)
                self.eventSet.removeAttachmentEvent(polyomino)
            del self.attachmentEventMap[coord]

    def RemoveAttachmentEventFromMap(self, coord, polyomino):
        if polyomino in self.attachmentEventMap[coord]:
            self.attachmentEventMap[coord].remove(polyomino)
            if len(self.attachmentEventMap[coord]) == 0:
                del self.attachmentEventMap[coord]

    def UpdateBindingStrength(self, polyomino):
        if polyomino in self.bindingStrengthMap:
            oldStrength = self.bindingStrengthMap[polyomino]
            strength = self.GetBindingStrength(polyomino)
            if strength != oldStrength:
                self.eventSet.removeDetachmentEvent(polyomino, oldStrength)
                self.eventSet.addDetachmentEvent(polyomino, strength)
                self.bindingStrengthMap[polyomino] = strength

    def GetBindingStrength(self, polyomino):
        strength = 0
        for coord in polyomino.glueMap:
            if coord in self.locationMap:
                other = self.locationMap[coord]
                if coord in other.domains:
                    for otherDomain in other.domains[coord]:
                        if otherDomain == polyomino.glueMap[coord]:
                            strength += otherDomain[2]
        return strength

    def GetFittingPolyominoes(self, coords):
        possible = set()
        for coord in coords:
            for polyominoType in self.attachableTypes:
                for block in polyominoType.blocks:
                    offset = coord - block
                    polyomino = Polyomino(polyominoType, offset)
                    if not self.OverlapsExistingPolyomino(polyomino):
                        if self.minBinding is None:
                            possible.add(polyomino)
                        elif self.GetBindingStrength(polyomino) >= self.minBinding:
                            possible.add(polyomino)
        return possible

    def Step(self, forwardRate, Gse, Gmc):
        #self.eventSet.Print(Gse, Gmc)
        event, dt = self.eventSet.sample(forwardRate, Gse, Gmc)
        if event.eventType == ATTACHMENT:
            self.AddPolyomino(event.polyomino)
        else:
            self.RemovePolyomino(event.polyomino)

        return dt

class PolyominoSystem:
    def __init__(self, filename):
        self.time = 0
        self.steps = 0
        self.ReadFromXML(filename)

    def Step(self):
        self.time += self.assembly.Step(self.forwardRate, self.Gse, self.Gmc)
        self.steps += 1

    def WriteToXML(self, filename):
        root = self.xmlTree.getroot()

        eltAssembly = root.find('assembly')
        if eltAssembly is None:
            eltAssembly = ET.SubElement(root, 'assembly')
            
        eltPolyominoes = eltAssembly.find('Polyominoes')
        if eltPolyominoes is None:
            eltPolyominoes = ET.SubElement(eltAssembly, 'Polyominoes')
        else:
            eltPolyominoes.clear()

        for polyomino in self.assembly.polyominoes:
            eltPolyomino = ET.SubElement(eltPolyominoes, 'Polyomino')
            eltPolyominoType = ET.SubElement(eltPolyomino, 'PolyominoType')
            eltTranslation = ET.SubElement(eltPolyomino, 'translation')

            eltPolyominoType.text = polyomino.polyominoType.name
            eltTranslation.text = str(polyomino.offset)

        self.xmlTree.write(filename)
        
            
    def ReadFromXML(self, filename):
        self.xmlTree = ET.parse(filename)
        
        root = self.xmlTree.getroot()
        if root.tag != 'PolyominoSystem':
            raise Exception('Invalid XML PolyominoSystem file')

        self.bindingThreshold = int(root.findtext('bindingThreshold', default='2'))
        self.Gse = float(root.findtext('Gse', default='10.0'))
        self.Gmc = float(root.findtext('Gmc', default='19.1'))
        self.forwardRate = float(root.findtext('forwardRate', default='5e8'))
        self.rotation = eval(root.findtext('rotation', default='True'))
        self.dimRestrictions = eval(root.findtext('dim_restrictions', default='None'))
        self.minBinding = int(root.findtext('min_binding', default='0'))

        self.polyominoTypes = {}
        for eltType in root.find('PolyominoTypes').findall('PolyominoType'):
            polyominoType = self.PolyominoTypeFromXML(eltType)
            if polyominoType.name in self.polyominoTypes:
                raise Exception(f'PolyominoType name collision: {polyominoType.name}')
            self.polyominoTypes[polyominoType.name] = polyominoType

        seedPolyominoes = set()
        eltSeed = root.find('seed')
        if eltSeed:
            for eltPolyomino in eltSeed.findall('./Polyominoes/Polyomino'):
                typeName = eltPolyomino.findtext('PolyominoType')
                if typeName not in self.polyominoTypes:
                    raise Exception(f'polyomino found with type invalid type {typeName}')
                polyominoType = self.polyominoTypes[typeName]
                translation = eval(eltPolyomino.findtext('translation', default='(0,0,0)'))
                offset = Coord(*translation)
                polyomino = Polyomino(polyominoType, offset)
                seedPolyominoes.add(polyomino)

        assemblyPolyominoes = set()
        eltAssembly = root.find('assembly')
        if eltAssembly:
            for eltPolyomino in eltAssembly.findall('./Polyominoes/Polyomino'):
                typeName = eltPolyomino.findtext('PolyominoType')
                if typeName not in self.polyominoTypes:
                    raise Exception(f'polyomino found with type invalid type {typeName}')
                polyominoType = self.polyominoTypes[typeName]
                translation = eval(eltPolyomino.findtext('translation', default='(0,0,0)'))
                offset = Coord(*translation)
                polyomino = Polyomino(polyominoType, offset)
                assemblyPolyominoes.set(polyomino)

        self.assembly = PolyominoAssembly(self.polyominoTypes, self.minBinding)

        for polyomino in seedPolyominoes:
            self.assembly.AddPolyomino(polyomino, True)
            
        for polyomino in assemblyPolyominoes.difference(seedPolyominoes):
            self.assembly.AddPolyomino(polyomino)
                                
    def PolyominoTypeFromXML(self, eltType):
        name = eltType.findtext('name')
        color = eltType.findtext('color')
        concentration = float(eltType.findtext('concentration', default='1.0'))
        placementHint = eval(eltType.findtext('placement_hint', default='None'))

        blocks = set()
        domains = {}
        
        for eltBlock in eltType.findall('.blocks/block'):
            coordTuple = eval(eltBlock.findtext('coords'))
            coord = Coord(*coordTuple)
            blocks.add(coord)

            domains[coord] = set()
            for eltDomain in eltBlock.findall('.domains/domain'):
                label = eltDomain.findtext('label')
                direction = eltDomain.findtext('direction')
                strength = int(eltDomain.findtext('strength', default='1'))
                domains[coord].add((label, direction, strength))

            if len(domains[coord]) == 0:
                del domains[coord]
                
        return PolyominoType(name, color, concentration, blocks, domains,
                             placementHint, self.dimRestrictions)


def GetConfigValue(conf_line, config):
    key = conf_line.partition('=')[0].strip()
    val = conf_line.partition('=')[2].strip()
    if val.isnumeric():
        val = int(val)
    else:
        try:
            val = float(val)
        except:
            pass
    config[key] = val

def ReadConfig(conf_file):
    config = dict()
    cFile = open(conf_file, "r")
    for line in cFile:
        line = line.partition('#')[0].strip() # ignore lines that begin with '#' or portions of lines after '#'
        if len(line) > 0:
            GetConfigValue(line, config)
    return config

def DisplayHelp():
    print("\nUsage:")
    print("-h, --help\t\t\tDisplays this help message")
    print("-s <file>, --sys=<file>\t\tInput polyomino system file (e.g. `test_system.xml')")
    print("-c <file>, --conf=<file>\tInput configuration file (e.g. 'test_conf.conf')")
    print("-o <prefix>, --out=<prefix>\tPrefix for output file name. If not specified 'out' is used (e.g. 'out.xml')")
    print("-d <dir>, --dir=<dir>\t\tDirectory for output files. If not specified, the current directory is used.")
    print("")
    exit()

if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "hs:c:o:d:",
                            ["help", "sys=", "conf=", "out=", "dir="])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()
        
    system_file = None
    conf_file   = None
    out_prefix  = 'out'
    out_dir     = ''
    
    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-s','--sys']:
            system_file = p
        elif o in ['-c', '--conf']:
            conf_file = p
        elif o in ['-o', '--out']:
            out_prefix = p
        elif o in ['-d', '--dir']:
            out_dir = p + "/"
        else:
            DisplayHelp()

    if system_file is None:
        print("ERROR: No input polyomino system file given.\n")
        DisplayHelp()
    else:
        try:
            system = PolyominoSystem(system_file)
        except Exception as e:
            print("\nERROR: Invalid input polyomino system file given: '" + system_file + "'\n")
            print("Exception: " + str(e))
            DisplayHelp()
        
    if conf_file is None:
        print("ERROR: No input configuration file given.\n")
        DisplayHelp()
    else:
        try:
            config = ReadConfig(conf_file)
        except Exception as e:
            print("\nERROR: Invalid input configuration file given: '" + conf_file + "'\n")
            print("Exception: " + str(e))
            DisplayHelp()
        
    if not 'num_steps' in config:
        print("ERROR: 'num_steps' parameter must be included in the configuration file")
        exit()
        
    if not 'output_interval' in config:
        print("ERROR: 'output_interval' parameter must be included in the configuration file")
        exit()
        
    num_simulations = 1
    if 'num_simulations' in config:
        num_simulations = config['num_simulations']
        
    print(f'\nLoaded configuration: {config}\n')

    for sim_num in range(1,num_simulations+1):
        print(f'Beginning simulation {sim_num} of {num_simulations}...')
        for i in range(config['num_steps']):
            system.Step()
            if (i % config['output_interval']) == 0:
                print(system.time)
        if num_simulations == 1:
            system.WriteToXML(f'{out_dir}{out_prefix}.xml')
        else:
            system.WriteToXML(f'{out_dir}{out_prefix}{sim_num}.xml')
        print(f'... simulation {sim_num} of {num_simulations} complete.')
    
    
